{# Import the theme's layout. #}
{% extends "!layout.html" %}

{% block htmltitle %}
    {% if title == '' or title == 'Home' %}
        <title>{{ docstitle|e }}</title>
    {% else %}
        <title>{{ title|striptags|e }}{{ titlesuffix }}</title>
    {% endif %}
{% endblock %}

{% block menu %}
    {{ super() }}

    {% if menu_links %}
        <p class="caption">
            <span class="caption-text">External links</span>
        </p>
        <ul>
            {% for text, link in menu_links %}
                <li class="toctree-l1"><a href="{{ link }}">{{ text }}</a></li>
            {% endfor %}
        </ul>
    {% endif %}
{% endblock %}
